package fos

import scala.util.parsing.input.Positional

enum Term extends Positional:
  case True
  case False
  case Zero
  case Succ(t: Term)
  case Pred(t: Term)
  case IsZero(t: Term)
  case If(cond: Term, t1: Term, t2: Term)
  case Var(name: String)
  case Abs(v: String, tp: TypeTree, t: Term)
  case App(t1: Term, t2: Term)
  case Let(x: String, tp: TypeTree, v: Term, t: Term)

  override def toString() = this match
    case True => "true"
    case False => "false"
    case Zero => "0"
    case Succ(t) => "succ " + t
    case Pred(t) => "pred " + t
    case IsZero(t) => "iszero " + t
    case If(cond, t1, t2) => "if " + cond + " then " + t1 + " else " + t2
    case Var(name) => name
    case Abs(v, tp, t) => "(\\" + v + ":" + tp + "." + t + ")"
    case App(t1, t2) => t1.toString + (t2 match
      case App(_, _) => " (" + t2.toString + ")" // left-associative
      case _         => " " + t2.toString
    )
    case Let(x, tp, v, t) =>
      "let " + x + ":" + tp + " = " + v + " in " + t

// Note that TypeTree is distinct from Type.
// The former is how types are parsed, the latter is how types are represented internally.
// We need this distinction because:
// 1) There are type vars, which can't be written by our users, but are needed by the inferencer.
// 2) There are empty types, which can be written, but aren't directly supported by the inferencer.
enum TypeTree extends Positional:
  case BoolTypeTree()
  case NatTypeTree()
  case FunTypeTree(t1: TypeTree, t2: TypeTree)
  case EmptyTypeTree()

  def tpe: Type = this match
    case BoolTypeTree() => Type.BoolType    
    case NatTypeTree() => Type.NatType
    case FunTypeTree(t1, t2) => Type.FunType(t1.tpe, t2.tpe)
    case EmptyTypeTree() =>
      throw new UnsupportedOperationException
    
    
  override def toString() = this match
    case BoolTypeTree() => "Bool"
    case NatTypeTree() => "Nat"
    case FunTypeTree(t1, t2) =>
      (t1 match
        case FunTypeTree(_, _) => "(" + t1 + ")" // right-associative
        case _ => t1.toString
      ) + "->" + t2
    case EmptyTypeTree() => "_"

